/*
 * vim: ts=4 nobackup
 */


#include "USB.H"
#include "USBDescriptor.H"
#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdlib.h>


using namespace std;
using namespace avelino;


const uint8_t USBSetupPacket::REQUEST_GET_DESCRIPTOR    = 0x06;
const uint8_t USBSetupPacket::REQUEST_SET_ADDRESS       = 0x05;
const uint8_t USBSetupPacket::REQUEST_GET_STATUS        = 0x00;
const uint8_t USBSetupPacket::REQUEST_GET_CONFIGURATION = 0x08;
const uint8_t USBSetupPacket::REQUEST_SET_CONFIGURATION = 0x09;
const uint8_t USB::STATUS_IDLE = 1;
const uint8_t USB::STATUS_HALT = 2;
const uint8_t USB::DEVICE_DESCRIPTOR        = 1;
const uint8_t USB::CONFIGURATION_DESCRIPTOR = 2;
uint8_t USB::whichDescriptor;
uint8_t *USB::toSend;
uint16_t USB::toSendSize;
uint8_t USB::status = 0;
uint8_t USB::buffer[64];
USBSetupPacket *USB::setupPacketReceived = NULL;
uint8_t USB::address = 0x00;
uint8_t USB::txBuffer[64];
uint8_t USB::rxBuffer[64];
bool USB::rxBufferFull;


void USB::powerOff() {
	PRR1 |= 1 << PRUSB;
}


void USB::done() {
	// detach USB interface
	UDCON |= (1 << DETACH);
	// disable USB interface
	USBCON &= ~(1 << USBE);
	// disable PLL
	PLLCSR |= (1 << PLLE);
	// disable USB pad regulator
	UHWCON &= ~(1 << UVREGE);
}


bool USB::configureEndpoint0() {
	UENUM = 0;   // select endpoint 0
	UECONX = (1 << EPEN);    // enable endpoint 0
	UECFG0X = 0;   // endpoint 0 is a control endpoint
	UECFG1X = (1 << EPSIZE1) | (1 << EPSIZE0) | (1 << ALLOC);    // endpoint 0: 64 bytes, one bank, allocate mem
	UEINTX = 0;
	// UEIENX = (1 << RXSTPE) | (1 << RXOUTE) | (1 << TXINE);
	UEIENX = (1 << RXSTPE);
	// check configuration
	//////return (UESTA0X & (1 << CFGOK));

	UENUM = 1;               // select endpoint 1
	UECONX = (1 << EPEN);    // enable endpoint 1
	UECFG0X = (1 << EPTYPE1) | (1 << EPDIR);    // endpoint 1 is a bulk in endpoint
	UECFG1X = (1 << EPSIZE1) | (1 << EPSIZE0) | (1 << ALLOC);    // endpoint 1: 64 bytes, one bank, allocate mem
	UEINTX = 0;
	UEIENX = (1 << TXINE);

	UENUM = 2;               // select endpoint 2
	UECONX = (1 << EPEN);    // enable endpoint 2
	UECFG0X = (1 << EPTYPE1);   // endpoint 2 is a bulk out endpoint
	UECFG1X = (1 << EPSIZE1) | (1 << EPSIZE0) | (1 << ALLOC);   // endpoint 2: 64 bytes, one bank, allocate mem
	UEINTX = 0;
	UEIENX = (1 << RXOUTE);

	USB::rxBufferFull = false;
}


void USB::init() {
	USB::setupPacketReceived = NULL;
	cli();
	// configure usb controller
	UHWCON = 0x01;
	// enable usb
	USBCON = (1 << USBE) | (1 << FRZCLK);
	// configure PLL
	PLLCSR = 0x12;
	PLLFRQ = (1 << PLLUSB) | (1 << PDIV3) | (1 << PDIV1) | (1 << PLLTM0);
	// wait for PLL lock
	while (!(PLLCSR & (1 << PLOCK)))
		;
	// start usb clock
	USBCON = (1 << USBE) | (1 << OTGPADE);
	// enable attach resistor
	UDCON = 0;
	// enable "end of reset" and "start of frame" device interrupts
	// UDIEN = (1 << EORSTE) | (1 << SOFE);
	UDIEN = (1 << EORSTE);
	USB::status = USB::STATUS_IDLE;
	USB::address = 0x00;
	sei();
}


ISR(USB_GEN_vect) {
	USB::__general_isr();
}


void USB::__general_isr() {
	if (UDINT & (1 << EORSTI))
		USB::configureEndpoint0();
	UDINT = 0;
}


ISR(USB_COM_vect) {
	USB::__endpoint_isr();
}


void USB::sendDataToEndpoint0() {
	while (USB::toSendSize > 0) {
		// wait for host ready for in packet
		while (!(UEINTX & ((1 << TXINI) | (1 << RXOUTI))))
			;
		if (UEINTX & (1 << RXOUTI))
			break;
		// send in packet
		uint16_t n = (USB::toSendSize >= 64) ? 64 : USB::toSendSize;
		USB::writeFromBuffer(n, USB::toSend);
		USB::toSend += n;
		USB::toSendSize -= n;
		UEINTX = ~(1 << TXINI);
	}
}


void USB::__endpoint_isr() {
	if (USB::status == USB::STATUS_IDLE) {
		UENUM = 0;
		uint8_t aux = UEINTX;
		if (aux & (1 << RXSTPI)) {
			// setup packet received
			USB::readOnBuffer(8, USB::buffer);
			USB::setupPacketReceived = (USBSetupPacket *) USB::buffer;
			UEINTX = ~((1 << RXSTPI) | (1 << RXOUTI) | (1 << TXINI));
			if (USB::setupPacketReceived->bRequest == USBSetupPacket::REQUEST_GET_DESCRIPTOR) {
				uint8_t descriptorType = USB::setupPacketReceived->wValue >> 8;
				USB::toSendSize = 0;
				if (descriptorType == 1) {
					// get the device descriptor
					USB::toSend = (uint8_t *) &MyUSBSerialDescriptorContainer.deviceDescriptor;
					USB::toSendSize = sizeof(USBDeviceDescriptor);
					USB::toSendSize = (USB::toSendSize > USB::setupPacketReceived->wLength) ? USB::setupPacketReceived->wLength : USB::toSendSize;
				}
				else if (descriptorType == 2) {
					// get the configuration descriptor
					USB::toSend = (uint8_t *) &MyUSBSerialDescriptorContainer.configurationDescriptor;
					USB::toSendSize = sizeof(USBSerialDescriptorContainer) - sizeof(USBDeviceDescriptor);
					USB::toSendSize = (USB::toSendSize > USB::setupPacketReceived->wLength) ? USB::setupPacketReceived->wLength : USB::toSendSize;
				}
				USB::sendDataToEndpoint0();
			}
			else if (USB::setupPacketReceived->bRequest == USBSetupPacket::REQUEST_SET_ADDRESS) {
				UEINTX = ~(1 << TXINI);
				USB::address = USB::setupPacketReceived->wValue;
				UDADDR = USB::address;
				while (!(UEINTX & (1 << TXINI)))
					;
				UDADDR |= (1 << ADDEN);
				UEINTX = ~(1 << TXINI);
			}
			else if (USB::setupPacketReceived->bRequest == USBSetupPacket::REQUEST_GET_STATUS) {
				USB::buffer[0] = 0;
				USB::buffer[1] = 0;
				USB::toSend = (uint8_t *) USB::buffer;
				USB::toSendSize = 2;
				USB::sendDataToEndpoint0();
			}
			else if (USB::setupPacketReceived->bRequest == USBSetupPacket::REQUEST_GET_CONFIGURATION) {
				USB::buffer[0] = 1;
				USB::toSend = (uint8_t *) USB::buffer;
				USB::toSendSize = 1;
				USB::sendDataToEndpoint0();
			}
			else if (USB::setupPacketReceived->bRequest == USBSetupPacket::REQUEST_SET_CONFIGURATION) {
				UEINTX = ~(1 << TXINI);
			}
		}
		// check for bulk out transfer
		UENUM = 2;
		if (UEINTX & (1 << RXOUTI)) {
			UEINTX &= ~(1 << RXOUTI);
			uint16_t numBytesReceived = (((uint16_t) UEBCHX) << 8) | ((uint16_t) UEBCLX);
			numBytesReceived = (numBytesReceived > 64) ? 64 : numBytesReceived;
			USB::readOnBuffer(numBytesReceived, USB::rxBuffer);
			USB::rxBufferFull = true;
			UEINTX &= ~(1 << FIFOCON);
		}
		// check for bulk in transfer
		UENUM = 1;
		if (UEINTX & (1 << TXINI)) {
			UEINTX &= ~(1 << TXINI);
			// TODO: write data to buffer
			UEINTX &= ~(1 << FIFOCON);
		}
	}
	else if (USB::status == USB::STATUS_HALT) {
		UENUM = 0;
		UEINTX = ~((1 << RXSTPI) | (1 << RXOUTI) | (1 << TXINI));
		UECONX = (1 << STALLRQ) | (1 << EPEN);   // stall
	}
}


void USB::readOnBuffer(uint16_t bytes, uint8_t *buffer) {
	uint8_t *p = buffer;
	while (bytes > 0) {
		*p = UEDATX;
		p++;
		bytes--;
	}
}


void USB::writeFromBuffer(uint16_t bytes, uint8_t *buffer) {
	uint8_t *p = buffer;
	while (bytes > 0) {
		UEDATX = *p;
		p++;
		bytes--;
	}
}


